from .maximum_mean_discrepancy import maximum_mean_discrepancy
from .root_mean_squared_error import root_mean_squared_error

from ...utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])
