"""
Provides :py:class:`~bayesflow.workflows.BasicWorkflow`, a high-level interface for working on typical BayesFlow
applications without having to worry about the internals.
"""

from .basic_workflow import BasicWorkflow

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])
